#include <iostream>  // For error reporting
#include <cctype>
#include <vector>
#include <map>

#include "syntaxparser.hh"

typedef unsigned LabelType;

static const LabelType FirstLabel = 0;

static map<string, LabelType> Variables;

static unsigned RunCommands(const string &target, unsigned ekapos=0)
{
    unsigned a=ekapos;
    for(;;)
    {
        unsigned b = target.find('=', a);
        if(b == target.npos)break;
        string varname = target.substr(a, b-a);
        
        a = b+1;

        LabelType label;
        const char *s = target.c_str() + a;
        for(label=0; isdigit(*s); ++s, ++a)label=label*10+(*s-'0');
        
        Variables[varname] = label;
        
    //  cerr << "Setting '" << varname << "' = " << label << endl;
        
        while(*s==',')++s,++a;
    }
    return a;    
}

class Choice
{
public:
    string text;
    string target;
public:
    Choice(const string &txt, const string &targ) : text(txt), target(targ)
    {
    }
public:
    inline const string &gettext() const
    {
        return text;
    }
    LabelType gettarget() const
    {
        unsigned a = RunCommands(target);
        if(isdigit(target[a]))
        {
            LabelType label;
            const char *s = target.c_str() + a;
            for(label=0; isdigit(*s); ++s)label=label*10+(*s-'0');
            return label;
        }
        return Variables[target];
    }
};

class ChoiceSet : public vector<Choice>
{
    string Commands;
public:
    void AddCommand(const string &s)
    {
        if(Commands.size())Commands += ',';
        Commands += s;
    }
    void Activate() const
    {
        RunCommands(Commands);
    }
};

class Labels : public map<LabelType, ChoiceSet>
{
};

Syntax::Syntax(const char *const s)
{
    Choices = new Labels();
    Load(s);
}
Syntax::~Syntax()
{
    delete Choices;
}

void Syntax::Load(const char *s)
{
    while(*s)
    {
        LabelType label;

        for(label=0; isdigit(*s); ++s)label=label*10+(*s-'0');

        ChoiceSet Setti;
        
        for(;;)
        {
            string text, target;
            while(*s==' '||*s=='\t')++s;
            if(*s=='#' && s[1]=='#')
            {
                for(s+=2; *s!='\n' && *s; ++s)text += *s;
                Setti.AddCommand(text);
                if(*s++ != '\n')
                {
                    cerr << "No '\\n'\n" << flush;
                    throw;
                }
            }            
            else
            {
                for(; *s!='\t'&&*s!=' '&&*s; ++s)text+=*s=='_'?' ':*s;
                while(*s==' '||*s=='\t')++s;
                for(; *s!='\n'&&*s; ++s)target += *s;
                if(*s++ != '\n')
                {
                    cerr << "No '\\n'\n" << flush;
                    throw;
                }
                Choice tmp(text, target);
                
                ChoiceSet::const_iterator i;
                for(i=Setti.begin(); i!=Setti.end(); ++i)
                {
                    if(i->gettext() == text.substr(0, i->gettext().size()))
                    {
                        cerr << "Ambiguous in " << label << ": \"" << i->gettext() << "\" and \"" << text << "\"\n" << flush;
                        throw;
                    }
                }
                
                Setti.push_back(tmp);
            }
            if(*s!=' ' && *s!='\t')break;
        }
        
        unsigned n = Setti.size();
        if((n & (n-1)) != 0)
        {
            cerr << "Entry " << label << " has non-2^ -count of choices\n" << flush;
            throw;
        }
        if(Choices->find(label) != Choices->end())
        {
            cerr << "Redefinition of label " << label << endl << flush;
            throw;
        }        
        (*Choices)[label] = Setti;
    }
}

int Syntax::operator+ (const string &s) const // pakkaus
{
    class bitti
    {
        int pos, sz;
        const unsigned char *data;
    public:
        bitti(const string &p) : pos(0), data((const unsigned char *)p.c_str()) { sz=p.size()*8; }
        inline int left() const { return sz-pos; }
        unsigned fetch(unsigned n)
        {
            unsigned res;
            unsigned bpos=0;
            for(res=0; n>0; ++pos,--n)
                res |= ((data[pos>>3] >> (pos&7)) & 1) << bpos++;
            return res;
        }
    } data(s);
    
    LabelType label = FirstLabel;
    
    result = "";
    
    while(data.left() > 0)
    {
        Labels::const_iterator i;
        i = Choices->find(label);
        if(i == Choices->end())
        {
            cerr << "Label " << label << " not found\n" << flush;
            throw;
        }
        unsigned b=0, c = i->second.size();
        while(c > unsigned(1 << b))++b;
        unsigned n = data.fetch(b);
        //  cerr << "Selection count = " << c << ", fetching " << b << " bits: " << n << endl << flush;
        
        const ChoiceSet &Setti = i->second;
        
        Setti.Activate();
        
        result += Setti[n].gettext();
        label = Setti[n].gettarget();
    }
    return 0;
}

int Syntax::operator- (const string &s) const // purku
{
    result = "";
    
    class bitti
    {
        unsigned pos;
        unsigned char data;
        string &result;
    public:
        bitti(string &kohde) : pos(0), data(0), result(kohde) { }
        void add(unsigned res, unsigned n)
        {
            unsigned bpos=0;
            for(; n>0; --n)
            {
                data |= ((res >> bpos++) & 1) << pos;
                if(++pos == 8)
                {
                    result += data;
                    data = 0;
                    pos = 0;
                }
            }
        }
        void finish() const
        {
            if(pos & 7)result += data;
        }
    } data(result);
    
    LabelType label = FirstLabel;

    for(unsigned srcpos=0; srcpos < s.size(); )
    {
        Labels::const_iterator i = Choices->find(label);
        if(i == Choices->end())
        {
            cerr << "Label " << label << " not found\n" << flush;
            throw;
        }
        
        i->second.Activate();
        
        unsigned a, b=0, c=i->second.size();
        while(c > unsigned(1 << b))++b;
        for(a=0; a<c; ++a)
        {
            const string &txt = i->second[a].gettext();
            unsigned txtsize = txt.size();
            if(s.substr(srcpos, txtsize) == txt)
            {
            //  cerr << "Max count = " << c << ", adding " << b << " bits: " << a << endl << flush;
                srcpos += txtsize;
                label = i->second[a].gettarget();
                break;
            }
        }
        if(a == c)
        {
            cerr << "Not found, at '" << s.substr(srcpos) << "'\n" << flush;
            throw;
        }
        data.add(a, b);
    }
    
    data.finish();
    
    return 0;
}
